#ifndef __WAKSMANNETWORK_HPP__
#define __WAKSMANNETWORK_HPP__

#include <vector>

/*
  WaksmanNetwork Class: Contains a Waksman permutation network and can apply it to an input array.
    setPermutation(uint32_t *permutation): Takes permutation as an array of N index
      values (i.e. values in [N]) and sets the Waksman network switches to that permutation.
    applyPermutation(unsigned char *buf): Takes buffer of N items of block_size bytes each and
      applies stored permutation in-place (i.e. modifying input buffer).
*/

class WaksmanNetwork {
  uint32_t N; // number of items to permute
  std::vector<bool> inSwitches; // input layer of switches
  std::vector<bool> outSwitches; // output layer of switches
  std::vector<WaksmanNetwork> subnetworks; // sub Waksman networks (zero or two)

  public:
  void setPermutation(const std:vector<uint32_t> &permutation);
  void applyPermutation(unsigned char *buf, size_t block_size);

  WaksmanNetwork(uint32_t N);
  WaksmanNetwork(const std::vector<uint32_t> &outside_wires, bool top_subnetwork);

  ~WaksmanNetwork();

  private:
  std::vector<uint32_t> wires; // wire indices
};

#endif
